/*
 * Format an HTML source in a consistent manner.
 *
 * Created 9 May 1998
 * Bert Bos <bert@w3.org>
 * $Id: normalize.c,v 1.9 1998/12/17 15:10:42 bbos Exp $
 */
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include "export.h"
#include "types.e"
#include "tree.e"
#include "html.e"
#include "scan.e"
#include "api.h"
#include "textwrap.e"

static Tree tree;
static Boolean do_xml = False;
static Boolean has_errors = False;


/* handle_error -- called when a parse error occurred */
void handle_error(void *clientdata, const unsigned char *s, int lineno)
{
  fprintf(stderr, "%d: %s\n", lineno, s);
  has_errors = True;
}

/* start -- called before the first event is reported */
void* start(void)
{
  tree = create();
  return NULL;
}
  
/* end -- called after the last event is reported */
void end(void *clientdata)
{
  /* skip */
}

/* handle_comment -- called after a comment is parsed */
void handle_comment(void *clientdata, unsigned char *commenttext)
{
  tree = append_comment(tree, commenttext);
}

/* handle_text -- called after a text chunk is parsed */
void handle_text(void *clientdata, unsigned char *text)
{
  tree = append_text(tree, text);
}

/* handle_declaration -- called after a declaration is parsed */
void handle_declaration(void *clientdata, unsigned char *decl)
{
  tree = append_declaration(tree, decl);
}

/* handle_proc_instr -- called after a PI is parsed */
void handle_proc_instr(void *clientdata, unsigned char *pi_text)
{
  tree = append_procins(tree, pi_text);
}

/* handle_starttag -- called after a start tag is parsed */
void handle_starttag(void *clientdata, unsigned char *name, pairlist attribs)
{
  tree = html_push(tree, name, attribs);
}

/* handle_emptytag -- called after an empty tag is parsed */
void handle_emptytag(void *clientdata, unsigned char *name, pairlist attribs)
{
  tree = html_push(tree, name, attribs);
}

/* handle_endtag -- called after an endtag is parsed (name may be "") */
void handle_endtag(void *clientdata, unsigned char *name)
{
  tree = html_pop(tree, name);
  free(name);
}

/* only_space -- check if s contains only whitespace */
static Boolean only_space(unsigned char *s)
{
  for (; isspace(*s); s++);
  return *s == '\0';
}

/* pp -- print the document normalized */
static void pp(Tree n, Boolean preformatted, Boolean allow_text)
{
  Boolean pre, mixed;
  unsigned char *s;
  pairlist h;
  size_t i;
  Tree l;

  switch (n->tp) {
    case Text:
      if (!allow_text) {
	assert(only_space(n->text));
      } else {
	s = n->text;
	i = strlen(s);
	if (s[0] == '\r' && s[1] == '\n') {s += 2; i -= 2;}
	else if (s[0] == '\n' || s[0] == '\r') {s++; i--;}
	if (i > 1 && s[i-1] == '\n' && s[i-2] == '\r') i -= 2;
	else if (i > 0 && (s[i-1] == '\r' || s[i-1] == '\n')) i--;
	outn(s, i, preformatted);
      }
      break;
    case Comment:
      out("<!--", True); out(n->text, True);
      if (allow_text || preformatted) out("-->", True);
      else outln("-->", preformatted);
      break;
    case Declaration:
      out("<!", False); out(n->text, True); outln(">", False);
      break;
    case Procins:
      out("<?", False); out(n->text, True);
      if (allow_text || preformatted) out(">", False);
      else outln(">", False);
      break;
    case Element:
      if (!preformatted && break_before(n->name)) outln(NULL, False);
      out("<", preformatted); out(n->name, preformatted);
      if (break_before(n->name)) inc_indent();
      for (h = n->attribs; h != NULL; h = h->next) {
	out(" ", preformatted); out(h->name, preformatted);
	if (h->value != NULL) {
	  out("=\"", preformatted);
	  out(h->value, preformatted);
	  outc('"', preformatted);
	} else if (do_xml) {
	  out("=\"", preformatted);
	  out(h->name, preformatted);
	  outc('"', preformatted);
	}
      }
      if (is_empty(n->name)) {
	assert(n->children == NULL);
	out(do_xml ? " />" : ">", True);
	if (break_before(n->name)) dec_indent();
	if (!preformatted && break_after(n->name)) outln(NULL, False);
      } else {
	out(">", preformatted);
	pre = is_pre(n->name);
	mixed = is_mixed(n->name);
	for (l = n->children; l != NULL; l = l->sister)
	  pp(l, pre, mixed);
	if (break_before(n->name)) dec_indent();
	if (do_xml || need_etag(n->name)) {
	  out("</", preformatted); out(n->name, preformatted);
	  out(">", preformatted);
	}
	if (!preformatted && break_after(n->name)) outbreak();
      }
      break;
    default:
      assert(!"Cannot happen");
  }
}

/* prettyprint -- print the tree normalized */
static void prettyprint(Tree t)
{
  Tree h;
  assert(t->tp == Root);
  for (h = t->children; h != NULL; h = h->sister) pp(h, False, False);
  flush();
}

/* usage -- print usage message and exit */
static void usage(unsigned char *prog)
{
  fprintf(stderr, "Usage: %s [-x] [-i indent] [-l linelen] [file]\n", prog);
  exit(1);
}

/* main -- main body */
int main(int argc, char *argv[])
{
  int c;

  while ((c = getopt(argc, argv, "xi:l:")) != -1)
    switch (c) {
      case 'x': do_xml = True; break;
      case 'i': set_indent(atoi(optarg)); break;
      case 'l': set_linelen(atoi(optarg)); break;
      default: usage(argv[0]);
    }
  if (optind == argc) yyin = stdin;
  else if (optind == argc - 1) yyin = fopen(argv[optind], "r");
  else usage(argv[0]);
  if (yyin == NULL) {perror(argv[1]); exit(2);}
  if (yyparse() != 0) {exit(3);}
  tree = get_root(tree);
  prettyprint(tree);
  return has_errors ? 1 : 0;
}
