/*
 * Parse HTML and output in approximate "nsgmls" format. The differences
 * are that comments are also printed (see * below), that implied
 * attributes are not, and that entities are left unexpanded. Use
 * "unent" to expand entities to UTF-8. Also \n is not used, instead
 * characters are printed like this:
 *
 *     &#10;   A line feed
 *
 *     &#13;   A carriage return
 *
 *     &#9;    A tab
 *
 * The possible command characters and arguments  are  as  follows:
 *
 *     (gi  The start of an element whose generic identifier is gi.
 *          Any  attributes  for this element will have been speci-
 *          fied with A commands.
 *
 *     )gi  The end of an element whose generic identifier is gi.
 *
 *     -data
 *          Data.
 *
 *     ?pi  A processing instruction with data pi.
 *
 *     *comment
 *          A comment
 *
 *     Aname val
 *          The next element to start has an  attribute  name  with
 *          value val which takes one of the following forms:
 *
 *          CDATA data
 *               The attribute is character data.  This is used for
 *               attributes whose declared value is CDATA.
 *
 *          ID token
 *               The attribute is an ID value.  This will be output
 *               only  if  the -oid option is specified.  Otherwise
 *               TOKEN will be used for ID values.
 *
 *	  Implied attribute are not shown.
 *
 *     Llineno
 *          Set the current line number.
 *          This will be output only if the -l option has
 *          been given.
 *
 *
 * Author: Bert Bos
 * Created: 2 Dec 1998
 * Version: $Id: pipe.c,v 1.3 1998/12/17 15:10:47 bbos Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "export.h"
#include "types.e"
#include "html.e"
#include "scan.e"
#include "api.h"

extern int yylineno;				/* From scan.l */

static Boolean has_error = False;
static Boolean in_text = False;
static Boolean linenumbering = False;


/* escape -- print a string with certain characters escaped */
static void escape(const unsigned char *s)
{
  for (; *s; s++)
    switch (*s) {
      case '\r': printf("&#13;"); break;
      case '\t': printf("&#9;"); break;
      case '\n': printf("&#10;"); break;
      default: putchar(*s);
    }
}

/* down -- convert a string to lowercase */
static inline void down(unsigned char *s)
{
  for (; *s; s++) *s = tolower(*s);
}

/* --------------- implements interface api.h -------------------------- */

/* handle_error -- called when a parse error occurred */
void handle_error(void *clientdata, const unsigned char *s, int lineno)
{
  fprintf(stderr, "%d: %s\n", lineno, s);
  has_error = True;
}

/* start -- called before the first event is reported */
void* start(void)
{
  return NULL;
}
  
/* end -- called after the last event is reported */
void end(void *clientdata)
{
  if (in_text) {putchar('\n'); in_text = False;}
}

/* handle_comment -- called after a comment is parsed */
void handle_comment(void *clientdata, unsigned char *commenttext)
{
  if (in_text) {putchar('\n'); in_text = False;}
  if (linenumbering) printf("L%d\n", yylineno);
  putchar('*');
  escape(commenttext);
  putchar('\n');
}

/* handle_text -- called after a text chunk is parsed */
void handle_text(void *clientdata, unsigned char *text)
{
  if (! in_text) {
    if (linenumbering) printf("L%d\n", yylineno);
    putchar('-');
    in_text = True;
  }
  escape(text);
}

/* handle_declaration -- called after a declaration is parsed */
void handle_declaration(void *clientdata, unsigned char *decl)
{
  /* ToDo: output this as well? */
}

/* handle_proc_instr -- called after a PI is parsed */
void handle_proc_instr(void *clientdata, unsigned char *pi_text)
{
  if (in_text) {putchar('\n'); in_text = False;}
  if (linenumbering) printf("L%d\n", yylineno);
  putchar('?');
  escape(pi_text);
  putchar('\n');
}

/* handle_starttag -- called after a start tag is parsed */
void handle_starttag(void *clientdata, unsigned char *name, pairlist attribs)
{
  pairlist p;

  if (in_text) {putchar('\n'); in_text = False;}
  for (p = attribs; p; p = p->next) {
    putchar('A');
    down(p->name);
    printf("%s", p->name);
    if (strcmp(p->name, "id") == 0) printf(" ID "); else printf(" CDATA ");
    if (p->value) escape(p->value); else printf("%s", p->name);
    putchar('\n');
  }
  if (linenumbering) printf("L%d\n", yylineno);
  putchar('(');
  printf("%s", name);
  putchar('\n');
}

/* handle_emptytag -- called after an empty tag is parsed */
void handle_emptytag(void *clientdata, unsigned char *name, pairlist attribs)
{
  pairlist p;

  if (in_text) {putchar('\n'); in_text = False;}
  for (p = attribs; p; p = p->next) {
    putchar('A');
    down(p->name);
    printf("%s", p->name);
    if (strcmp(p->name, "id") == 0) printf(" ID "); else printf(" CDATA ");
    if (p->value) escape(p->value); else printf("%s", p->name);
    putchar('\n');
  }
  if (linenumbering) printf("L%d\n", yylineno);
  putchar('(');
  printf("%s", name);
  putchar('\n');
  putchar(')');
  printf("%s", name);
  putchar('\n');
}

/* handle_endtag -- called after an endtag is parsed (name may be "") */
void handle_endtag(void *clientdata, unsigned char *name)
{
  if (in_text) {putchar('\n'); in_text = False;}
  if (linenumbering) printf("L%d\n", yylineno);
  putchar(')');
  printf("%s", name);
  putchar('\n');
}

/* --------------------------------------------------------------------- */

/* usage -- print usage message and exit */
static void usage(unsigned char *prog)
{
  fprintf(stderr, "Usage: %s [-l] [html-file]\n", prog);
  exit(2);
}

int main(int argc, char *argv[])
{
  int i;

  for (i = 1; i < argc && argv[i][0] == '-' && !eq(argv[i], "--"); i++) {
    switch (argv[i][1]) {
      case 'l': linenumbering = True;; break;
      default: usage(argv[0]);
    }
  }
  if (i < argc && eq(argv[i], "--")) i++;

  if (i == argc) yyin = stdin;
  else if (i == argc - 1) yyin = fopen(argv[i], "r");
  else usage(argv[0]);

  if (yyin == NULL) {perror(argv[i]); exit(1);}

  if (yyparse() != 0) exit(3);

  return has_error ? 1 : 0;
}
