/*
 * Insert an active ToC between "<!--begin-toc-->" and "<!--end-toc-->",
 * or replacing the comment "<!--toc-->"
 *
 * Headers with class "no-toc" will not be listed in the ToC.
 *
 * The ToC links to elements with ID attributes as well as with
 * empty <A NAME> elements.
 *
 * Bert Bos <bert@w3.org>
 * Created Sep 1997
 * $Id: toc.c,v 1.9 1998/12/17 15:10:50 bbos Exp $
 */
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "export.h"
#include "types.e"
#include "tree.e"
#include "html.e"
#include "scan.e"

extern long random(void);
extern void srandom(unsigned seed);

#define BEGIN_TOC "begin-toc"			/* <!--begin-toc--> */
#define END_TOC "end-toc"			/* <!--end-toc--> */
#define TOC "toc"				/* <!--toc--> */
#define NO_TOC "no-toc"				/* CLASS="... no-toc..." */

#define WRITE True
#define NO_WRITE False
#define EXPAND True
#define NO_EXPAND False
#define KEEP_ANCHORS True
#define REMOVE_ANCHORS False

static Tree tree;
static int toc_low = 1, toc_high = 6;		/* Which headers to include */


/* handle_error -- called when a parse error occurred */
EXPORT void handle_error(void *clientdata, const unsigned char *s, int lineno)
{
  fprintf(stderr, "%d: %s\n", lineno, s);
}

/* start -- called before the first event is reported */
EXPORT void* start(void)
{
  tree = create();
  return NULL;
}
  
/* end -- called after the last event is reported */
EXPORT void end(void *clientdata)
{
  /* skip */
}

/* handle_comment -- called after a comment is parsed */
EXPORT void handle_comment(void *clientdata, unsigned char *commenttext)
{
  tree = append_comment(tree, commenttext);
}

/* handle_text -- called after a tex chunk is parsed */
EXPORT void handle_text(void *clientdata, unsigned char *text)
{
  tree = append_text(tree, text);
}

/* handle_declaration -- called after a declaration is parsed */
EXPORT void handle_declaration(void *clientdata, unsigned char *decl)
{
  tree = append_declaration(tree, decl);
}

/* handle_proc_instr -- called after a PI is parsed */
EXPORT void handle_proc_instr(void *clientdata, unsigned char *pi_text)
{
  tree = append_procins(tree, pi_text);
}

/* handle_starttag -- called after a start tag is parsed */
EXPORT void handle_starttag(void *clientdata, unsigned char *name, pairlist attribs)
{
  tree = html_push(tree, name, attribs);
}

/* handle_emptytag -- called after an empty tag is parsed */
EXPORT void handle_emptytag(void *clientdata, unsigned char *name, pairlist attribs)
{
  tree = html_push(tree, name, attribs);
}

/* handle_pop -- called after an endtag is parsed (name may be "") */
EXPORT void handle_endtag(void *clientdata, unsigned char *name)
{
  tree = html_pop(tree, name);
}

/* contains -- check if string contains a certain word, return pointer */
static unsigned char* contains(unsigned char *s, unsigned char *word)
{
  unsigned char c, *t;
  t = strstr(s, word);
  if (t == NULL)
    return NULL;				/* Not found */
  if (t != s && !isspace(*(t - 1)))
    return NULL;				/* Not beginning of word */
  if ((c = *(t + strlen(word))) != '\0' && !isspace(c))
    return NULL;				/* Not end of word */
  return t;
}

/* Check for class=toc in list of attributes */
static Boolean class_contains_toc(pairlist attribs)
{
  pairlist p;

  for (p = attribs; p != NULL; p = p->next) {
    if (strcasecmp(p->name, "class") == 0 && contains(p->value, TOC))
      return True;
  }
  return False;
}

static void expand(Tree t, Boolean write, Boolean exp, Boolean keep_anchors);

/* toc -- create a table of contents */
static void toc(Tree t, int *curlevel)
{
  int level;
  Tree h;
  unsigned char s[20], *id, *val;

  switch (t->tp) {
    case Text: break;
    case Comment: break;
    case Declaration: break;
    case Procins: break;
    case Element:
      /* Check if the element is a heading */
      if (eq(t->name, "h1")) level = 1;
      else if (eq(t->name, "h2")) level = 2;
      else if (eq(t->name, "h3")) level = 3;
      else if (eq(t->name, "h4")) level = 4;
      else if (eq(t->name, "h5")) level = 5;
      else if (eq(t->name, "h6")) level = 6;
      else level = -1;
      /* Check if it has a "no-toc" class */
      if (level > 0 && get_attrib(t, "class", &val) && contains(val, NO_TOC))
	level = -1;
      /* If it's a header for the ToC, create a list item for it */
      if (level >= toc_low && level <= toc_high) {
	/* Ensure there is an ID to point to */
	if (! get_attrib(t, "id", NULL)) {
	  sprintf(s, "id%011lu", random());
	  set_attrib(t, "id", s);
	}
	while (*curlevel > level) {
	  printf("</ul>\n");
	  (*curlevel)--;
	}
	while (*curlevel < level - 1) {
	  printf("<li><ul>\n");
	  (*curlevel)++;
	}
	if (*curlevel == level - 1) {
	  printf("<ul>\n");
	  (*curlevel)++;
	}
	(void) get_attrib(t, "id", &id);
	printf("<li><a href=\"#%s\">", id);
	expand(t, WRITE, NO_EXPAND, REMOVE_ANCHORS);
	printf("</a>\n");
      } else {
	for (h = t->children; h != NULL; h = h->sister) toc(h, curlevel);
      }
      break;
    case Root:
      for (h = t->children; h != NULL; h = h->sister) toc(h, curlevel);
      break;
    default: assert(! "Cannot happen");
  }
}

/* expand -- write the tree, inserting ID's at H* and inserting a toc */
static void expand(Tree t, Boolean write, Boolean exp, Boolean keep_anchors)
{
  Tree h;
  unsigned char s[20];
  pairlist a;
  int level;

  for (h = t->children; h != NULL; h = h->sister) {
    switch (h->tp) {
      case Text:
	if (write) printf("%s", h->text);
	break;
      case Comment:
	if (exp && (eq(h->text, TOC) || eq(h->text, BEGIN_TOC))) {
	  printf("<!--%s-->\n", BEGIN_TOC);
	  printf("<ul class=\"%s\">\n", TOC);
	  level = toc_low;
	  toc(get_root(t), &level);
	  while (level >= toc_low) {
	    printf("</ul>\n");
	    level--;
	  }
	  printf("<!--%s-->", END_TOC);
	  if (eq(h->text, BEGIN_TOC))
	    write = False;			/* Suppress old ToC */
	} else if (exp && eq(h->text, END_TOC)) {
	  write = True;
	} else {
	  printf("<!--%s-->", h->text);
	}
	break;
      case Declaration:
	if (write) printf("<!%s>", h->text);
	break;
      case Procins:
	if (write) printf("<?%s>", h->text);
	break;
      case Element:
	if (eq(h->name, "h1") || eq(h->name, "h2") || eq(h->name, "h3")
	    || eq(h->name, "h4") || eq(h->name, "h5") || eq(h->name, "h6")) {
	  /* Give headers an ID, if they don't have one */
	  if (! get_attrib(h, "id", NULL)) {
	    sprintf(s, "id%011lu", random());
	    set_attrib(h, "id", s);
	  }
	}
	if (write) {
	  if (! keep_anchors && eq(h->name, "a")) {
	    /* Don't write the <a> and </a> tags */
	    expand(h, write, exp, False);
	  } else if (eq(h->name, "a") && class_contains_toc(h->attribs)) {
	    /* This <a> was inserted by toc itself; remove it */
	    expand(h, write, exp, False);
	  } else {
	    printf("<%s", h->name);
	    for (a = h->attribs; a != NULL; a = a->next) {
	      printf(" %s", a->name);
	      if (a->value != NULL) printf("=\"%s\"", a->value);
	    }
	    if (is_empty(h->name)) {
	      assert(h->children == NULL);
	      printf(" />");
	    } else {
	      unsigned char *val;
	      printf(">");
	      /* Insert an <A NAME> if element has an ID */
	      if (is_mixed(h->name) && get_attrib(h, "id", &val))
		printf("<a class=\"%s\" name=\"%s\"></a>", TOC, val);
	      expand(h, write, exp, keep_anchors);
	      printf("</%s>", h->name);
	    }
	  }
	}
	break;
      case Root:
	assert(! "Cannot happen");
	break;
      default:
	assert(! "Cannot happen");
    }
  }
}

/* usage -- print usage message and exit */
static void usage(unsigned char *name)
{
  fprintf(stderr, "Usage: %s [-l low] [-h high] [html-file]\n", name);
  exit(1);
}


int main(int argc, char *argv[])
{
  int i;

  yyin = stdin;
  for (i = 1; i < argc; i++) {
    if (eq(argv[i], "-l")) {
      if (i >= argc - 1) usage(argv[0]);
      toc_low = atoi(argv[++i]);
    } else if (eq(argv[i], "-h")) {
      if (i >= argc - 1) usage(argv[0]);
      toc_high = atoi(argv[++i]);
    } else {
      yyin = fopen(argv[i], "r");
      if (yyin == NULL) {
	perror(argv[1]);
	exit(2);
      }
    }
  }
  if (toc_low < 1) toc_low = 1;
  if (toc_high > 6) toc_high = 6;

  if (yyparse() != 0) {
    exit(3);
  }
  tree = get_root(tree);
  srandom(time(NULL));
  expand(tree, WRITE, EXPAND, KEEP_ANCHORS);
  return 0;
}
