/* $Id: tree.c,v 1.12 1998/12/05 00:52:52 bbos Exp $ */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "export.h"
#include "types.e"
#include "dtd.e"

EXPORT typedef enum {
  Element, Text, Comment, Declaration, Procins, Root
} Nodetype;

EXPORT typedef struct _node {
  Nodetype tp;
  unsigned char *name;
  pairlist attribs;
  unsigned char *text;
  struct _node *parent;
  struct _node *sister;
  struct _node *children;
} Node, *Tree;

  

/* create -- create an empty tree */
EXPORT Tree create(void)
{
  Tree t = malloc(sizeof(*t));
  assert(t != NULL);
  t->tp = Root;
  t->name = "";
  t->children = t->children = NULL;
  return t;
}

/* get_root -- return root of tree */
EXPORT Tree get_root(Tree t)
{
  while (t->tp != Root) t = t->parent;
  return t;
}

/* get_attrib -- return a ptr to the value of a named attibute, or False */
EXPORT Boolean get_attrib(Node *e, unsigned char *attname, unsigned char **val)
{
  pairlist h;

  assert(e->tp == Element);
  for (h = e->attribs; h != NULL; h = h->next) {
    if (eq(h->name, attname)) {
      if (val != NULL) *val = h->value;
      return True;
    }
  }
  return False;
}

/* set_attrib -- set an attribute to a value */
EXPORT void set_attrib(Node *e, unsigned char *name, unsigned char *value)
{
  pairlist h, p;
  assert(e->tp == Element);
  for (h = e->attribs; h != NULL; h = h->next) {
    if (eq(h->name, name)) {
      free(h->value);
      h->value = strdup(value);
      return;
    }
  }
  h = malloc(sizeof(*h));
  assert(h != NULL);
  h->name = strdup(name);
  h->value = strdup(value);
  h->next = NULL;
  /* Insert into sorted list */
  if (e->attribs == NULL || strcmp(h->name, e->attribs->name) < 0) {
    h->next = e->attribs;
    e->attribs = h;
  } else {
    p = e->attribs;
    while (p->next != NULL && strcmp(h->name, p->next->name) > 0) p = p->next;
    h->next = p->next;
    p->next = h;
  }
}

/* push -- add a child node to the tree */
static Tree push(Tree t, Node *n)
{
  if (t->children == NULL) {
    t->children = n;
  } else {
    Tree h = t->children;
    while (h->sister != NULL) h = h->sister;
    h->sister = n;
  }
  n->parent = t;
  return n;
}

/* pop -- go up one level */
static Tree pop(Tree t)
{
  assert(t != NULL);
  assert(t->tp != Root);
  return t->parent;
}

/* append -- add at end of children */
static Tree append(Tree t, Node *n)
{
  assert(t != NULL);
  if (t->children == NULL) {
    t->children = n;
  } else {
    Tree h = t->children;
    while (h->sister != NULL) h = h->sister;
    h->sister = n;
  }
  n->parent = t;
  return t;
}

/* is_known -- true if the element is an HTML 4 element */
EXPORT Boolean is_known(unsigned char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return info != NULL;
}

/* is_pre -- true if the element has preformatted content */
EXPORT Boolean is_pre(unsigned char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return info && info->pre;
}

/* need_etag -- true if the element's end tag is required */
EXPORT Boolean need_etag(unsigned char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return !info || info->etag;
}

/* is_empty -- true if element is empty */
EXPORT Boolean is_empty(unsigned char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return info && info->empty;
}

/* has_parent -- true if c accepts p as a child */
static Boolean has_parent(const char *c, const char *p)
{
  const ElementType *info = lookup_element(c, strlen(c));
  int i;
  if (!info) return False;
  for (i = 0; info->parents[i]; i++)
    if (eq(info->parents[i], p)) return True;
  return False;
}

/* preferred_parent -- return first possible parent of e */
static char *preferred_parent(const char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  assert(info != NULL);				/* element is known */
  assert(info->parents[0] != NULL);		/* element is not root */
  return info->parents[0];
}
  
/* is_root -- true if e has no possible parents */
static Boolean is_root(const char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  assert(info != NULL);				/* element is known */
  return info->parents[0] == NULL;
}
  
/* is_mixed -- true if e accepts text content */
EXPORT Boolean is_mixed(const char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return !info || info->mixed;
}
  
/* break_before -- true if element looks better with a newline before it */
EXPORT Boolean break_before(const char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return info && info->break_before;
}
  
/* break_after -- true if element looks better with a newline after it */
EXPORT Boolean break_after(const char *e)
{
  const ElementType *info = lookup_element(e, strlen(e));
  return info && info->break_after;
}
  
/* html_push -- add an element to the tree, open or close missing elements */
EXPORT Tree html_push(Tree t, unsigned char *elem, pairlist attr)
{
  Node *n = malloc(sizeof(*n));
  assert(n != NULL);
  n->tp = Element;
  n->name = elem;
  n->attribs = attr;
  n->sister = n->children = NULL;

  /* Unknown elements are just pushed where they are */
  if (!is_known(elem)) return push(t, n);

  if (is_root(elem)) {
    while (t->tp != Root) t = pop(t);		/* Make sure root is at root */
  } else {
    /* Close omitted end tags */
    while (t->tp != Root && !has_parent(elem, t->name) && !need_etag(t->name))
      t = pop(t);
    /* If no valid parent, fabricate one */
    if (t->tp == Root || (is_known(t->name) && !has_parent(elem, t->name)))
      t = html_push(t, preferred_parent(elem), NULL);
  }
  t = push(t, n);

#if 0
  /* Insert omitted tags */
  if (t->tp == Root && !eq(elem, "html"))
    t = html_push(t, "html", NULL);
  if (in_block(elem)) {
    /* pop until in a block */
    while (!eq(t->name, "html") && !is_block(t->name)) t = pop(t);
    if (!is_block(t->name)) t = html_push(t, "body", NULL);
    t = push(t, n);
  } else if (in_text(elem)) {
    /* insert a p if needed */
    if (!is_text(t->name) && is_known(t->name)) t = html_push(t, "p", NULL);
    t = push(t, n);
  } else if (in_head(elem)) {
    while (!eq(t->name, "html") && !eq(t->name, "head")) t = pop(t);
    if (!eq(t->name, "head")) t = html_push(t, "head", NULL);
    t = push(t, n);
  } else if (eq(elem, "dt") || eq(elem, "dd")) {
    /* pop until we're in a dl */
    while (!eq(t->name, "html") && !eq(t->name, "body") && !eq(t->name, "dl"))
      t = pop(t);
    if (!eq(t->name, "dl")) t = html_push(t, "dl", NULL);
    t = push(t, n);
  } else if (eq(elem, "li")) {
    /* pop until we're in a list */
    while (!eq(t->name, "html") && !eq(t->name, "body")
	   && !eq(t->name, "ol") && !eq(t->name, "ul")) t = pop(t);
    if (!eq(t->name, "ol") && !eq(t->name, "ul")) t = html_push(t, "ul", NULL);
    t = push(t, n);
  } else if (eq(elem, "option")) {
    /* pop until we're in a select */
    while (!eq(t->name, "html") && !eq(t->name, "body")
	   && !eq(t->name, "select")) t = pop(t);
    if (!eq(t->name, "select")) t = html_push(t, "select", NULL);
    t = push(t, n);
  } else if (eq(elem, "thead") || eq(elem, "tfoot") || eq(elem, "tbody")
	     || eq(elem, "colgroup")) {
    /* pop until we're in a table */
    while (!eq(t->name, "html") && !eq(t->name, "body")
	   && !eq(t->name, "table")) t = pop(t);
    if (!eq(t->name, "table")) t = html_push(t, "table", NULL);
    t = push(t, n);
  } else if (eq(elem, "tr")) {
    /* pop until we're in a table or rowgroup */
    while (!eq(t->name, "html") && !eq(t->name, "table")
	   && !eq(t->name, "thead") && !eq(t->name, "tfoot")
	   && !eq(t->name, "tbody") && !eq(t->name, "body")) t = pop(t);
    if (!eq(t->name, "table") && !eq(t->name, "thead")
	&& !eq(t->name, "tfoot") && !eq(t->name, "tbody"))
      t = html_push(t, "table", NULL);
    t = push(t, n);
  } else if (eq(elem, "th") || eq(elem, "td")) {
    /* pop until we're in a tr */
    while (!eq(t->name, "html") && !eq(t->name, "tr") && !eq(t->name, "body"))
      t = pop(t);
    if (! eq(t->name, "tr")) t = html_push(t, "tr", NULL);
    t = push(t, n);
  } else if (eq(elem, "body")) {
    while (!eq(t->name, "html")) t = pop(t);
    t = push(t, n);
  } else if (eq(elem, "head")) {
    while (!eq(t->name, "html")) t = pop(t);
    t = push(t, n);
  } else if (eq(elem, "html") && t->tp != Root) {
    ; /* ignore */
  } else /*if (is_known(elem))*/ {
    t = push(t, n);
  }
#endif
  if (is_empty(elem)) t = pop(t);
  return t;
}

/* html_pop -- close an open element */
EXPORT Tree html_pop(Tree t, unsigned char *elem)
{
  Tree h = t;
  assert(t != NULL);
  if (*elem == '\0') {				/* </> */
    if (t->tp != Root) t = pop(t);
  } else {					/* </name> */
    for (h = t; h->tp != Root && !eq(h->name, elem); h = h->parent) ;
    if (h->tp != Root) {			/* Found open element */
      while (t != h) t = pop(t);
      t = pop(t);
    }
  }
  return t;
}

/* append_comment -- add a comment to the tree */
EXPORT Tree append_comment(Tree t, unsigned char *comment)
{
  Node *n = malloc(sizeof(*n));
  assert(n != NULL);
  n->tp = Comment;
  n->text = comment;
  n->sister = n->children = NULL;
  return append(t, n);
}

/* append_declaration -- add a declaration to the tree */
EXPORT Tree append_declaration(Tree t, unsigned char *declaration)
{
  Node *n = malloc(sizeof(*n));
  assert(n != NULL);
  n->tp = Declaration;
  n->text = declaration;
  n->sister = n->children = NULL;
  return append(t, n);
}

/* append_procins -- append a processing instruction */
EXPORT Tree append_procins(Tree t, unsigned char *procins)
{
  Node *n = malloc(sizeof(*n));
  assert(n != NULL);
  n->tp = Procins;
  n->text = procins;
  n->sister = n->children = NULL;
  return append(t, n);
}

/* is_whitespace -- true if whole string consists of whitespace */
static Boolean is_whitespace(unsigned char *s)
{
  for (;*s != '\0'; s++)
    if (*s != ' ' && *s != '\t' && *s != '\r' && *s != '\n') return False;
  return True;
}

/* append_text -- append a text chunk to the document tree */
EXPORT Tree append_text(Tree t, unsigned char *text)
{
  Node *n;

  if (!is_whitespace(text) || (t->tp != Root && is_mixed(t->name))) {
    /* Close omitted end tags */
    while (t->tp != Root && !is_mixed(t->name) && !need_etag(t->name))
      t = pop(t);
    /* Fabricate a parent if needed */
    if (!is_mixed(t->name)) t = html_push(t, preferred_parent("%data"), NULL);
  }
  n = malloc(sizeof(*n));
  assert(n != NULL);
  n->tp = Text;
  n->text = text;
  assert(n->text != NULL);
  n->sister = n->children = NULL;
  return append(t, n);
}

static void dump2(Tree n)
{
  pairlist h;
  Tree l;

  switch (n->tp) {
    case Text: printf("%s", n->text); break;
    case Comment: printf("<!--%s-->", n->text); break;
    case Declaration: printf("<!%s>", n->text); break;
    case Procins: printf("<?%s>", n->text); break;
    case Element:
      printf("<%s", n->name);
      for (h = n->attribs; h != NULL; h = h->next) {
	printf(" %s", h->name);
	if (h->value != NULL) printf("=\"%s\"", h->value);
      }
      if (is_empty(n->name)) {
	assert(n->children == NULL);
	printf(" />");
      } else {
	printf(">");
	for (l = n->children; l != NULL; l = l->sister) dump2(l);
	printf("</%s>", n->name);
      }
      break;
    default:
      assert(!"Cannot happen");
  }
}

/* dumptree -- write out the tree below t (t's children, not t itself)*/
EXPORT void dumptree(Tree t)
{
  Tree h;

  printf("<?XML encoding=\"ISO8859-1\"?>\n");
  for (h = t->children; h != NULL; h = h->sister) dump2(h);
}
